/* -*-c-*-
 *
 * Pyke: the Python Kit for Extensions
 *
 * (c) 2019 Straylight/Edgeware
 */

/*----- Licensing notice --------------------------------------------------*
 *
 * This file is part of Pyke: the Python Kit for Extensions.
 *
 * Pyke is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * Pyke is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Pyke.  If not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/*----- Header files ------------------------------------------------------*/

#include "pyke.h"

/*----- External variables ------------------------------------------------*/

PyObject *modname;
PyObject *home_module;

/*----- Conversions -------------------------------------------------------*/

PyObject *getulong(unsigned long w)
{
  if (w <= LONG_MAX) return (PyInt_FromLong(w));
  else return (PyLong_FromUnsignedLong(w));
}

PyObject *getbool(int b)
  { if (b) RETURN_TRUE; else RETURN_FALSE; }

int convulong(PyObject *o, void *pp)
{
  unsigned long *p = pp;
  PyObject *t;

  if (!o) VALERR("can't delete");
#ifdef PY2
  if (PyInt_Check(o)) {
    long i = PyInt_AS_LONG(o);
    if (i < 0) VALERR("must be nonnegative");
    *p = i;
  } else
#endif
  {
    if ((t = PyNumber_Long(o)) == 0) goto end;
    *p = PyLong_AsUnsignedLong(t);
    Py_DECREF(t);
    if (PyErr_Occurred()) goto end;
  }
  return (1);
end:
  return (0);
}

int convuint(PyObject *o, void *pp)
{
  unsigned long u;
  unsigned *p = pp;

  if (!convulong(o, &u)) goto end;
  if (u > UINT_MAX) VALERR("out of range");
  *p = u;
  return (1);
end:
  return (0);
}

int convszt(PyObject *o, void *pp)
{
  unsigned long u;
  size_t *p = pp;

  if (!convulong(o, &u)) goto end;
  if (u > ~(size_t)0) VALERR("out of range");
  *p = u;
  return (1);
end:
  return (0);
}

int convbool(PyObject *o, void *pp)
{
  if (!o) VALERR("can't delete");
  *(int *)pp = PyObject_IsTrue(o);
  return (1);
end:
  return (0);
}

int convbin(PyObject *o, void *pp)
{
  struct bin *r = pp;

  if (BIN_CHECK(o)) {
    r->p = BIN_PTR(o);
    r->sz = BIN_LEN(o);
    return (1);
  }
#ifdef PY2
  if (PyUnicode_Check(o)) {
    o = _PyUnicode_AsDefaultEncodedString(o, 0);
    if (!o) return (0);
    r->p = PyString_AS_STRING(o);
    r->sz = PyString_GET_SIZE(o);
    return (1);
  }
#endif
  return (PyObject_AsReadBuffer(o, &r->p, &r->sz) ? 0 : 1);
}

/*----- Miscellaneous utilities -------------------------------------------*/

PyObject *abstract_pynew(PyTypeObject *ty, PyObject *arg, PyObject *kw)
{
  PyErr_SetString(PyExc_TypeError, "can't instantiate this class");
  return (0);
}

PyObject *enrich_compare(int op, int cmp)
{
  int r = -1;

  switch (op) {
    case Py_LT: r = cmp <  0; break;
    case Py_LE: r = cmp <= 0; break;
    case Py_EQ: r = cmp == 0; break;
    case Py_NE: r = cmp != 0; break;
    case Py_GE: r = cmp >= 0; break;
    case Py_GT: r = cmp >  0; break;
    default: assert(0);
  }
  return (getbool(r));
}

/*----- Saving and restoring exceptions ----------------------------------*/

void report_lost_exception_v(struct excinfo *exc,
			     const char *why, va_list ap)
{
  PyObject *hookfn = 0;
  PyObject *whyobj = 0;
  PyObject *obj = 0;

  /* Make sure we start out without a pending exception, or this will get
   * really confusing.
   */
  assert(!PyErr_Occurred());

  /* Format the explanation. */
  if (why) whyobj = TEXT_VFORMAT(why, ap);
  else { whyobj = Py_None; Py_INCREF(whyobj); }

  /* Find our home module's `lostexchook' function.  This won't work if
   * there's no module, or the function isn't defined, or it's `None'.
   */
  if (!home_module) goto sys;
  hookfn = PyObject_GetAttrString(home_module, "lostexchook");
  if (hookfn == Py_None) goto sys;
  else if (hookfn) ;
  else if (!PyErr_ExceptionMatches(PyExc_AttributeError)) goto ouch;
  else { PyErr_Clear(); goto sys; }

  /* Call the hook function. */
  obj = PyObject_CallFunction(hookfn, "(OOOO)",
			      whyobj, exc->ty, exc->val, exc->tb);
  if (!obj) goto ouch;
  goto end;

  /* Something went wrong reporting the problem. */
ouch:
  PySys_WriteStderr("\n!!! FAILURE REPORTING LOST EXCEPTION\n");
  PyErr_Print();
  /* drop through... */

  /* There was no hook, so try to do something sensible using
   * `sys.excepthook'.
   */
sys:
  PySys_WriteStderr("\n!!! LOST EXCEPTION: %s\n", TEXT_PTR(whyobj));
  RESTORE_EXCINFO(exc);
  PyErr_Print();
  /* drop through... */

  /* Clean up afterwards. */
end:
  Py_XDECREF(hookfn);
  Py_XDECREF(whyobj);
  Py_XDECREF(obj);
}

void report_lost_exception(struct excinfo *exc, const char *why, ...)
{
  va_list ap;

  va_start(ap, why);
  report_lost_exception_v(exc, why, ap);
  va_end(ap);
}

void stash_exception(struct excinfo *exc, const char *why, ...)
{
  va_list ap;
  struct excinfo stash;

  if (!exc->ty)
    STASH_EXCINFO(exc);
  else {
    va_start(ap, why);
    STASH_EXCINFO(&stash);
    report_lost_exception_v(&stash, why, ap);
    va_end(ap);
  }
}

void restore_exception(struct excinfo *exc, const char *why, ...)
{
  va_list ap;
  struct excinfo stash;

  if (!PyErr_Occurred())
    RESTORE_EXCINFO(exc);
  else {
    va_start(ap, why);
    STASH_EXCINFO(&stash);
    report_lost_exception_v(exc, why, ap);
    RESTORE_EXCINFO(&stash);
    va_end(ap);
  }
}

/*----- Type definitions --------------------------------------------------*/

static const PyTypeObject emptytype = { 0 };

void *newtype(PyTypeObject *metaty,
	      const PyTypeObject *skel,
	      const char *name)
{
  PyHeapTypeObject *ty =
    (PyHeapTypeObject *)_PyObject_GC_Malloc(_PyObject_VAR_SIZE(metaty, 0));
  if (!skel) skel = &emptytype;
  memcpy(ty, skel, sizeof(*skel));
#define COPY(blah) do {							\
    if (ty->ht_type.tp_as_##blah) {					\
      memcpy(&ty->as_##blah,						\
	     ty->ht_type.tp_as_##blah,					\
	     sizeof(ty->as_##blah));					\
      ty->ht_type.tp_as_##blah = &ty->as_##blah;			\
    }									\
  } while (0)
  COPY(number);
  COPY(sequence);
  COPY(mapping);
  COPY(buffer);
#undef COPY
  if (name)
    ty->ht_name = TEXT_FROMSTR(name);
  else if (ty->ht_type.tp_name)
    ty->ht_name = TEXT_FROMSTR(ty->ht_type.tp_name);
  else
    ty->ht_name = 0;
  if (ty->ht_name)
    ty->ht_type.tp_name = TEXT_STR(ty->ht_name);
  ty->ht_slots = 0;
#ifdef PY3
  ty->ht_qualname = 0;
#endif
  (void)PyObject_INIT(&ty->ht_type, metaty);
  Py_INCREF(metaty);
  return (ty);
}

void typeready(PyTypeObject *ty)
{
#ifdef PY3
  PyHeapTypeObject *hty = (PyHeapTypeObject *)ty;
  hty->ht_qualname = hty->ht_name;
#endif
  PyType_Ready(ty);
  PyDict_SetItemString(ty->tp_dict, "__module__", modname);
}

PyTypeObject *inittype(const PyTypeObject *tyskel,
		       PyTypeObject *base, PyTypeObject *meta)
{
  PyTypeObject *ty = newtype(meta, tyskel, 0);
  if (base) { ty->tp_base = base; Py_INCREF(base); }
  ty->tp_flags |= Py_TPFLAGS_HEAPTYPE;
  typeready(ty);
  return (ty);
}

/*----- Populating modules ------------------------------------------------*/

PyObject *mkexc(PyObject *mod, PyObject *base,
		const char *name, const PyMethodDef *mm)
{
  PyObject *nameobj = 0;
  PyObject *dict = 0;
  PyObject *exc = 0;
  PyObject *func = 0;
  PyObject *meth = 0;

  if ((dict = PyDict_New()) == 0) goto fail;

  if (mm) {
    while (mm->ml_name) {
      if ((func = PyCFunction_NewEx((/*unconst*/ PyMethodDef *)mm,
				    0, mod)) == 0 ||
	  (meth = PY23(PyMethod_New(func, 0, exc),
		       PyInstanceMethod_New(func))) == 0 ||
	  PyDict_SetItemString(dict, mm->ml_name, meth))
	goto fail;
      Py_DECREF(func); func = 0;
      Py_DECREF(meth); meth = 0;
      mm++;
    }
  }

  if ((nameobj = TEXT_FORMAT("%s.%s", PyModule_GetName(mod), name)) == 0 ||
      (exc = PyErr_NewException(TEXT_STR(nameobj), base, dict)) == 0)
    goto fail;

done:
  Py_XDECREF(nameobj);
  Py_XDECREF(dict);
  return (exc);

fail:
  Py_XDECREF(exc);
  Py_XDECREF(func);
  Py_XDECREF(meth);
  exc = 0;
  goto done;
}

void setconstants(PyObject *mod, const struct nameval *c)
{
  PyObject *x;
  unsigned long u;

  while (c->name) {
    u = c->value;
    if (u <= LONG_MAX) x = PyInt_FromLong(u);
    else if (c->f&CF_SIGNED) x = PyInt_FromLong(-1 - (long)(ULONG_MAX - u));
    else x = PyLong_FromUnsignedLong(u);
    PyModule_AddObject(mod, (/*unconst*/ char *)c->name, x);
    c++;
  }
}

/*----- Submodules --------------------------------------------------------*/

static PyMethodDef *global_methods;
static size_t nmethods, methodsz;

void addmethods(const PyMethodDef *m)
{
  size_t n, want, newsz;

  for (n = 0; m[n].ml_name; n++);
  want = nmethods + n + 1;
  if (want > methodsz) {
    newsz = methodsz ? 2*methodsz : 16;
    while (want > newsz) newsz *= 2;
    if (!global_methods)
      global_methods = PyObject_Malloc(newsz*sizeof(PyMethodDef));
    else
      global_methods = PyObject_Realloc(global_methods,
					newsz*sizeof(PyMethodDef));
    assert(global_methods);
    methodsz = newsz;
  }
  memcpy(global_methods + nmethods, m, n*sizeof(PyMethodDef));
  nmethods += n;
  global_methods[nmethods].ml_name = 0;
}

PyMethodDef *donemethods(void) { return (global_methods); }

/*----- Low-level Python interface ----------------------------------------*/

static PyObject *meth__set_home_module(PyObject *me, PyObject *arg)
{
  PyObject *mod;

  if (!PyArg_ParseTuple(arg, "O!:_set_home_module", &PyModule_Type, &mod))
    return (0);
  Py_XDECREF(home_module); home_module = mod; Py_INCREF(home_module);
  RETURN_NONE;
}

static const PyMethodDef methods[] = {
#define METHNAME(func) meth_##func
  METH	(_set_home_module, "_set_home_module(MOD)")
#undef METHNAME
  { 0 }
};

void pyke_core_pyinit(void) { addmethods(methods); }
void pyke_core_pyinsert(PyObject *mod) { ; }

/*----- That's all, folks -------------------------------------------------*/
